﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Net;
using System.Net.Http;
using System.IO;
using System.Net.Http.Headers;
using System.Security.Cryptography.X509Certificates;

namespace OpenSourceCSharpReptile
{
    public class EasyHTTP
    {

        #region 预定义方变量
        //默认的编码
        private Encoding _encoding = Encoding.Default;
        //Post数据编码
        private Encoding _postencoding = Encoding.Default;
        //HttpWebRequest对象用来发起请求
        private HttpClient _client = null;

        HttpClientHandler _handler = null;
        //设置head头数据
        private HttpRequestMessage _requestHeads = null;
        //这些头数据 不能加在head 中 得加在content.head中
        private static List<string> _list = new List<string>() { "Content-Disposition".ToLower(), "Content-Encoding".ToLower(), "Content-Language".ToLower(), "Content-Location".ToLower(), "Content-MD5".ToLower(), "Content-Range".ToLower(), "Expires".ToLower(), "LastModified".ToLower() };

        public async Task<HttpResult> GetHtml(HttpItem item)
        {
            string html = string.Empty;
            try
            {

                SetRequest(item);

                try
                {

                    var response = await _client.SendAsync(_requestHeads);

                    if (response.IsSuccessStatusCode)
                    {
                        var result = new HttpResult();
                        result.StatusCode = response.StatusCode;


                        #region base
                        //获取StatusCode

                        //获取StatusDescription
                        result.IsSuccessStatusCode = response.IsSuccessStatusCode;
                        //获取Headers
                        result.Header = response.Content.Headers;


                        //获取set-cookie
                        if (response.Headers.Contains("Set-Cookie"))
                        {
                            var cookies = response.Headers.GetValues("Set-Cookie");
                            string reCook = string.Empty;

                            if (item.AutoHanderCookie)
                            {
                                if (cookies != null) reCook = CookieHelper.GetSmallCookie(cookies);

                            }
                            else
                            {
                                reCook = string.Join(";", cookies);
                            }


                            if (item.AutoFixCookie)
                            {
                                result.Cookie = CookieHelper.UpdateMergeCookie(item.Cookie, reCook);
                            }
                            else
                            {
                                result.Cookie = reCook;
                            }



                        }


                        #endregion
                        //获取跳转地址
                        if (response.Headers.Location != null)
                        {
                            result.RedirectUrl = response.Headers.Location.ToString();
                        }

                        if (item.ResultType == ResultType.Byte)
                        {
                            result.ResultByte = await response.Content.ReadAsByteArrayAsync();
                        }
                        else
                        {
                            using (Stream myResponseStream = await response.Content.ReadAsStreamAsync())
                            {
                                using (StreamReader myStreamReader = new StreamReader(myResponseStream, _encoding))
                                {
                                    result.Html = myStreamReader.ReadToEnd();

                                }

                            }


                        }



                        return result;


                    }



                    return new HttpResult() { Cookie = string.Empty, Header = null, Html = "", StatusCode = response.StatusCode, StatusDescription = "服务器响应：" + response.StatusCode };


                }
                catch (Exception ex)
                {
                    return new HttpResult() { Cookie = string.Empty, Header = null, Html = ex.Message, StatusDescription = "请求时出错：" + ex.Message };
                }



            }
            catch (Exception ex)
            {

                //配置参数时出错
                return new HttpResult() { Cookie = string.Empty, Header = null, Html = ex.Message, StatusDescription = "配置参数时出错：" + ex.Message };
            }
        }




        /// <summary>
        /// 为请求准备参数
        /// </summary>
        ///<param name="item">参数列表</param>
        private void SetRequest(HttpItem item)
        {

            try
            {
                _handler = new HttpClientHandler();//调用代理的时候 需要

                if (item.URL.StartsWith("https")) System.Net.ServicePointManager.SecurityProtocol = SecurityProtocolType.Tls;

                _requestHeads = new HttpRequestMessage(item.Method, item.URL);


                //设置压缩 解压 方式
                _handler.AutomaticDecompression = item.Decompression;

                //Authorization 认证
                if (!string.IsNullOrEmpty(item.Authorization))
                {
                    _requestHeads.Headers.Authorization = new AuthenticationHeaderValue("Basic", item.Authorization);
                }

                //设置Post数据
                if (item.Method != HttpMethod.Get)
                {
                    SetPostData(item);
                }

                //设置Header参数
                if (item.Header != null && item.Header.Count > 0)
                {
                    foreach (string key in item.Header.AllKeys)
                    {
                        if (_list.Contains(key.ToLower()))//这些头不能加在请求头里面 得加在 content.head 里面
                        {
                            if (_requestHeads.Content != null)
                            {
                                _requestHeads.Content.Headers.TryAddWithoutValidation(key, item.Header[key]);
                            }


                        }
                        else
                        {

                            _requestHeads.Headers.Add(key, item.Header[key]);
                        }

                    }
                }
                // 设置代理
                SetProxy(item);
                if (item.ProtocolVersion != null) _requestHeads.Version = item.ProtocolVersion;


                if (!string.IsNullOrWhiteSpace(item.Host))
                {
                    _requestHeads.Headers.Host = item.Host;
                }

                //Accept
                //_requestHeads.Headers.Accept.Add(new System.Net.Http.Headers.MediaTypeWithQualityHeaderValue(item.Accept));

                _requestHeads.Headers.Add("Accept", item.Accept.Trim());

                if (item.KeepAlive)
                {
                    _requestHeads.Headers.ConnectionClose = false;
                }
                else
                {
                    _requestHeads.Headers.ConnectionClose = true;//非持久连接
                }



                //UserAgent客户端的访问类型，包括浏览器版本和操作系统信息
                _requestHeads.Headers.Add("User-Agent", item.UserAgent);
                // 编码

                if (item.Encoding != null)
                {
                    _encoding = item.Encoding;
                }

                //设置安全凭证
                _handler.Credentials = item.ICredentials;
                //设置Cookie
                SetCookie(item);
                //来源地址


                if (!string.IsNullOrEmpty(item.Referer)) _requestHeads.Headers.Referrer = new Uri(item.Referer);


                //是否执行跳转功能
                _handler.AllowAutoRedirect = item.Allowautoredirect;
                if (item.MaximumAutomaticRedirections > 0)
                {
                    _handler.MaxAutomaticRedirections = item.MaximumAutomaticRedirections;
                }



                //最后设置 clinet
                _client = new HttpClient(_handler);
                _client.Timeout = TimeSpan.FromSeconds(item.Timeout);//秒


            }
            catch (Exception ex)
            {
                Console.WriteLine(ex.Message);
                throw new Exception(ex.Message);
            }
        }




        /// <summary>
        /// 设置Cookie
        /// </summary>
        /// <param name="item">Http参数</param>
        private void SetCookie(HttpItem item)
        {
            if (!string.IsNullOrEmpty(item.Cookie))
            {
                _handler.UseCookies = false;
                _requestHeads.Headers.Add("Cookie", item.Cookie);
            }

        }
        /// <summary>
        /// 设置Post数据
        /// </summary>
        /// <param name="item">Http参数</param>
        private void SetPostData(HttpItem item)
        {

            //验证在得到结果时是否有传入数据

            if (item.PostEncoding != null)
            {
                _postencoding = item.PostEncoding;
            }
            byte[] buffer = null;
            //写入Byte类型
            if (item.PostDataType == PostDataType.Byte && item.PostdataByte != null && item.PostdataByte.Length > 0)
            {
                //验证在得到结果时是否有传入数据
                buffer = item.PostdataByte;
            }//写入文件
            else if (item.PostDataType == PostDataType.FilePath && !string.IsNullOrWhiteSpace(item.Postdata))
            {
                StreamReader r = new StreamReader(item.Postdata, _postencoding);
                buffer = _postencoding.GetBytes(r.ReadToEnd());
                r.Close();
            } //写入字符串
            else if (!string.IsNullOrWhiteSpace(item.Postdata))
            {
                buffer = _postencoding.GetBytes(item.Postdata);
            }
            if (buffer != null)
            {


                _requestHeads.Content = new ByteArrayContent(buffer, 0, buffer.Length);
                _requestHeads.Content.Headers.Add("Content-Length", buffer.Length.ToString());

            }


            if (_requestHeads.Content != null)
            {
                _requestHeads.Content.Headers.ContentType = new System.Net.Http.Headers.MediaTypeHeaderValue(item.ContentType);
            }



        }
        /// <summary>
        /// 设置代理
        /// </summary>
        /// <param name="item">参数对象</param>
        private void SetProxy(HttpItem item)
        {
            bool isIeProxy = false;
            if (!string.IsNullOrWhiteSpace(item.ProxyIp))
            {
                isIeProxy = item.ProxyIp.ToLower().Contains("ieproxy");
            }
            if (!string.IsNullOrWhiteSpace(item.ProxyIp) && !isIeProxy)
            {
                //设置代理服务器
                if (item.ProxyIp.Contains(":"))
                {
                    string[] plist = item.ProxyIp.Split(':');
                    WebProxy myProxy = new WebProxy(plist[0].Trim(), Convert.ToInt32(plist[1].Trim()));
                    //建议连接
                    myProxy.Credentials = new NetworkCredential(item.ProxyUserName, item.ProxyPwd);
                    //给当前请求对象
                    _handler.UseProxy = true;
                    _handler.Proxy = myProxy;
                }
                else
                {
                    WebProxy myProxy = new WebProxy(item.ProxyIp, false);
                    //建议连接
                    myProxy.Credentials = new NetworkCredential(item.ProxyUserName, item.ProxyPwd);
                    //给当前请求对象
                    _handler.UseProxy = true;
                    _handler.Proxy = myProxy;
                }
            }
            else if (isIeProxy)
            {
                //设置为IE代理
                _handler.UseProxy = true;// defaultProxy 	
                                         // 使用默认（系统）代理时，获取或设置要提交到默认代理服务器进行身份验证的凭据。 只有在 UseProxy 设置为 true 且 Proxy 设置为 null 时才使用默认代理。
            }
            else
            {
                _handler.Proxy = item.WebProxy;
            }
        }

        #endregion end class

        /// <summary>
        ///释放资源
        /// </summary>
        ~EasyHTTP()
        {

            if (_handler != null)
            {
                _handler.Dispose();
            }

            if (_requestHeads != null)
            {
                _requestHeads.Dispose();
            }


            if (_client != null)
            {
                _client.Dispose();
            }

        }

    }

    #region public calss
    /// <summary>
    /// Http请求参考类
    /// </summary>
    public class HttpItem
    {
        /// <summary>
        /// 处理程序使用的自动解压缩方法 Gzip Deflate none=不使用压缩 默认为none
        /// </summary>
        public DecompressionMethods Decompression = DecompressionMethods.None;

        public Task<HttpResult> GetHtml()
        {
            return new EasyHTTP().GetHtml(this);
        }

        /// <summary>
        /// Authorization 认证   Convert.ToBase64String(Encoding.UTF8.GetBytes($"{Username}:{Password}"))
        /// </summary>
        public string Authorization { get; set; }
        /// <summary>
        /// 请求URL必须填写
        /// </summary>
        public string URL { get; set; }
        HttpMethod _Method = HttpMethod.Get;
        /// <summary>
        /// 请求方式默认为GET方式,当为POST方式时必须设置Postdata的值
        /// </summary>
        public HttpMethod Method
        {
            get { return _Method; }
            set { _Method = value; }
        }
        int _Timeout = 100000;
        /// <summary>
        /// 默认请求超时时间 单位为秒
        /// </summary>
        public int Timeout
        {
            get { return _Timeout; }
            set { _Timeout = value; }
        }
        int _ReadWriteTimeout = 30000;
        /// <summary>
        /// 默认写入Post数据超时间
        /// </summary>
        public int ReadWriteTimeout
        {
            get { return _ReadWriteTimeout; }
            set { _ReadWriteTimeout = value; }
        }
        /// <summary>
        /// 设置Host的标头信息
        /// </summary>
        public string Host { get; set; }
        Boolean _KeepAlive = true;
        /// <summary>
        ///  获取或设置一个值，该值指示是否与 Internet 资源建立持久性连接默认为true。
        /// </summary>
        public Boolean KeepAlive
        {
            get { return _KeepAlive; }
            set { _KeepAlive = value; }
        }
        string _Accept = "*/*";
        /// <summary>
        /// 请求标头值 默认为text/html, application/xhtml+xml, */*
        /// </summary>
        public string Accept
        {
            get { return _Accept; }
            set { _Accept = value; }
        }
        string _ContentType = "application/x-www-form-urlencoded";
        /// <summary>
        /// 请求返回类型默认 text/html
        /// </summary>
        public string ContentType
        {
            get { return _ContentType; }
            set { _ContentType = value; }
        }
        string _UserAgent = "Mozilla/5.0 (Windows NT 6.1; WOW64; rv:17.0) Gecko/20100101 Firefox/17.0";
        /// <summary>
        /// 客户端访问信息默认Mozilla/5.0 (compatible; MSIE 9.0; Windows NT 6.1; Trident/5.0)
        /// </summary>
        public string UserAgent
        {
            get { return _UserAgent; }
            set { _UserAgent = value; }
        }
        /// <summary>
        /// 返回数据编码默认为NUll 一般为utf-8,gbk,gb2312
        /// </summary>
        public Encoding Encoding { get; set; }
        private PostDataType _PostDataType = PostDataType.String;
        /// <summary>
        /// Post的数据类型
        /// </summary>
        public PostDataType PostDataType
        {
            get { return _PostDataType; }
            set { _PostDataType = value; }
        }
        /// <summary>
        /// Post请求时要发送的字符串Post数据
        /// </summary>
        public string Postdata { get; set; }
        /// <summary>
        /// Post请求时要发送的Byte类型的Post数据
        /// </summary>
        public byte[] PostdataByte { get; set; }
        /// <summary>
        /// Cookie对象集合
        /// </summary>
        public CookieContainer CookieContainer { get; set; }
        /// <summary>
        /// 请求时的Cookie
        /// </summary>
        public string Cookie { get; set; }
        /// <summary>
        /// 来源地址，上次访问地址
        /// </summary>
        public string Referer { get; set; }
        /// <summary>
        /// 证书绝对路径
        /// </summary>
        public string CerPath { get; set; }
        /// <summary>
        /// 设置代理对象，不想使用IE默认配置就设置为Null，而且不要设置ProxyIp
        /// </summary>
        public WebProxy WebProxy { get; set; }
        private Boolean isToLower = false;
        /// <summary>
        /// 是否设置为全文小写，默认为不转化
        /// </summary>
        public Boolean IsToLower
        {
            get { return isToLower; }
            set { isToLower = value; }
        }
        private Boolean allowautoredirect = true;
        /// <summary>
        /// 支持跳转页面，查询结果将是跳转后的页面，默认是跳转
        /// </summary>
        public Boolean Allowautoredirect
        {
            get { return allowautoredirect; }
            set { allowautoredirect = value; }
        }
        private int connectionlimit = 1024;
        /// <summary>
        /// 最大连接数
        /// </summary>
        public int Connectionlimit
        {
            get { return connectionlimit; }
            set { connectionlimit = value; }
        }
        /// <summary>
        /// 代理Proxy 服务器用户名
        /// </summary>
        public string ProxyUserName { get; set; }
        /// <summary>
        /// 代理 服务器密码
        /// </summary>
        public string ProxyPwd { get; set; }
        /// <summary>
        /// 代理 服务IP,如果要使用IE代理就设置为ieproxy
        /// </summary>
        public string ProxyIp { get; set; }
        private ResultType resulttype = ResultType.String;
        /// <summary>
        /// 设置返回类型String和Byte
        /// </summary>
        public ResultType ResultType
        {
            get { return resulttype; }
            set { resulttype = value; }
        }
        private WebHeaderCollection header = new WebHeaderCollection();
        /// <summary>
        /// header对象
        /// </summary>
        public WebHeaderCollection Header
        {
            get { return header; }
            set { header = value; }
        }
        /// <summary>
        //     获取或设置用于请求的 HTTP 版本。返回结果:用于请求的 HTTP 版本。默认为 System.Net.HttpVersion.Version11。
        /// </summary>
        public Version ProtocolVersion { get; set; }
        private Boolean _expect100continue = false;
        /// <summary>
        ///  获取或设置一个 System.Boolean 值，该值确定是否使用 100-Continue 行为。如果 POST 请求需要 100-Continue 响应，则为 true；否则为 false。默认值为 true。
        /// </summary>
        public Boolean Expect100Continue
        {
            get { return _expect100continue; }
            set { _expect100continue = value; }
        }
        /// <summary>
        /// 设置509证书集合
        /// </summary>
        public X509CertificateCollection ClentCertificates { get; set; }
        /// <summary>
        /// 设置或获取Post参数编码,默认的为Default编码
        /// </summary>
        public Encoding PostEncoding { get; set; }
        private ResultCookieType _ResultCookieType = ResultCookieType.String;
        /// <summary>
        /// Cookie返回类型,默认的是只返回字符串类型
        /// </summary>
        public ResultCookieType ResultCookieType
        {
            get { return _ResultCookieType; }
            set { _ResultCookieType = value; }
        }
        private ICredentials _ICredentials = CredentialCache.DefaultCredentials;
        /// <summary>
        /// 获取或设置请求的身份验证信息。
        /// </summary>
        public ICredentials ICredentials
        {
            get { return _ICredentials; }
            set { _ICredentials = value; }
        }
        /// <summary>
        /// 设置请求将跟随的重定向的最大数目
        /// </summary>
        public int MaximumAutomaticRedirections { get; set; }
        private DateTime? _IfModifiedSince = null;
        /// <summary>
        /// 获取和设置IfModifiedSince，默认为当前日期和时间
        /// </summary>
        public DateTime? IfModifiedSince
        {
            get { return _IfModifiedSince; }
            set { _IfModifiedSince = value; }
        }




        private bool _auToFixCookie = true;
        /// <summary>
        /// 是否自动合并并更新Cookie 默认为true
        /// </summary>
        public bool AutoFixCookie
        {
            get { return _auToFixCookie; }
            set { _auToFixCookie = value; }
        }

        private bool _autoHandlerCookie = true;
        /// <summary>
        /// 是否自动处理无效Cookie 默认为true
        /// </summary>
        public bool AutoHanderCookie
        {
            get { return _autoHandlerCookie; }
            set { _autoHandlerCookie = value; }
        }

    }
    /// <summary>
    /// Http返回参数类
    /// </summary>
    public class HttpResult
    {
        /// <summary>
        /// Http请求返回的Cookie
        /// </summary>
        public string Cookie { get; set; }


        private string _html = string.Empty;
        /// <summary>
        /// 返回的String类型数据 只有ResultType.String时才返回数据，其它情况为空
        /// </summary>
        public string Html
        {
            get { return _html; }
            set { _html = value; }
        }
        /// <summary>
        /// 返回的Byte数组 只有ResultType.Byte时才返回数据，其它情况为空
        /// </summary>
        public byte[] ResultByte { get; set; }
        /// <summary>
        /// header对象
        /// </summary>
        public HttpContentHeaders Header { get; set; }
        /// <summary>
        ///http 是否响应成功
        /// </summary>
        public bool IsSuccessStatusCode { get; set; }

        /// <summary>
        /// http响应 描述信息
        /// </summary>
        public string StatusDescription { get; set; }

        /// <summary>
        /// 返回状态码,默认为OK
        /// </summary>
        public HttpStatusCode StatusCode { get; set; }
        /// <summary>
        /// 最后访问的URl
        /// </summary>
        public string ResponseUri { get; set; }
        /// <summary>
        /// 获取重定向的URl
        /// </summary>
        public string RedirectUrl
        {
            get; set;
        }
    }
    /// <summary>
    /// 返回类型
    /// </summary>
    public enum ResultType
    {
        /// <summary>
        /// 表示只返回字符串 只有Html有数据
        /// </summary>
        String,
        /// <summary>
        /// 表示返回字符串和字节流 ResultByte和Html都有数据返回
        /// </summary>
        Byte
    }
    /// <summary>
    /// Post的数据格式默认为string
    /// </summary>
    public enum PostDataType
    {
        /// <summary>
        /// 字符串类型，这时编码Encoding可不设置
        /// </summary>
        String,
        /// <summary>
        /// Byte类型，需要设置PostdataByte参数的值编码Encoding可设置为空
        /// </summary>
        Byte,
        /// <summary>
        /// 传文件，Postdata必须设置为文件的绝对路径，必须设置Encoding的值
        /// </summary>
        FilePath
    }
    /// <summary>
    /// Cookie返回类型
    /// </summary>
    public enum ResultCookieType
    {
        /// <summary>
        /// 只返回字符串类型的Cookie
        /// </summary>
        String,
        /// <summary>
        /// CookieCollection格式的Cookie集合同时也返回String类型的cookie
        /// </summary>
        CookieCollection
    }
    #endregion
}
